/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  (c) XNeur Team 2007
 *
 */
 
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef WITH_OPENAL

#include <stdlib.h>
#include <AL/al.h>
#include <AL/alc.h>
#include <AL/alut.h>

#include "xnconfig.h"
#include "xnconfig_files.h"

#include "log.h"
#include "types.h"
	
extern struct _xneur_config *xconfig;

void sound_init(void)
{
	alutInit(NULL, NULL);
}

void play_file(int file_type)
{
	if (xconfig->sound_mode != SOUND_ENABLED)
		return;

	char *path = get_file_path_name(SOUNDDIR, xconfig->sounds[file_type]);
	int need_free = TRUE;

	if (path == NULL)
	{
		path = xconfig->sounds[file_type];
		need_free = FALSE;
	}

	log_message(DEBUG, "Play sound sample %s", path);

	ALuint AlutBuffer;
	AlutBuffer = alutCreateBufferFromFile(path);

	if (need_free == TRUE)
		free(path);

	if (!AlutBuffer)
		return;

	ALuint AlutSource;
	alGenSources(1, &AlutSource);
	alSourcei(AlutSource, AL_BUFFER, AlutBuffer);
	alSourcePlay(AlutSource);

	do
		alDeleteSources(1, &AlutSource);
	while (alGetError() != AL_NO_ERROR);

	do
		alDeleteBuffers(1, &AlutBuffer);
	while (alGetError() != AL_NO_ERROR);

}

void sound_uninit(void)
{
	alutExit();
}

#else

void sound_init(void)
{}

void play_file(int file_type)
{
	file_type = file_type;
}

void sound_uninit(void)
{}

#endif
